{ *****************************************************
                   TCustomPopupEdit Component

  A set of components to simplify the use of popup dialogs. 
  These components basically act like super combo boxes.

                  Paul Warren
         HomeGrown Software Development
       (c) 1997 Langley British Columbia.
                (604) 856-6523
         e-mail:  hg_soft@uniserve.com
    Home page: http://users.uniserve.com/~hg_soft
  ***************************************************** }

unit Popedits;

interface

uses
  SysUtils, WinTypes, WinProcs, Messages, Classes, Graphics, Controls,
  Forms, Dialogs, StdCtrls, Mask, Buttons, DB, DBTables, Popups,
  DBCtrls;

type
  TCustomPopupEdit = class(TCustomMaskEdit)
  private
    { private declarations }
    FButton: TSpeedButton;
    FTextMargin: Integer;
    FKeepAspect: boolean;
    FAlignment: TAlignment;
    FFocused: Boolean;
    FOnBtnClick: TNotifyEvent;
    function GetGlyph: TBitMap;
    function GetNumGlyphs: integer;
    procedure SetGlyph(AGlyph: TBitmap);
    procedure SetNumGlyphs(Value: integer);
    procedure SetOnBtnClick(Value: TNotifyEvent);
    procedure CalcTextMargin;
    function GetMinHeight: Integer;
    procedure SetEditRect;
    procedure WMSize(var Message: TWMSize); message WM_SIZE;
  protected
    { protected declarations }
    procedure BtnClick(Sender: TObject); virtual;
    procedure CreateParams(var Params: TCreateParams); override;
    procedure CreateWnd; override;
  public
    { public declarations }
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    property Button: TSpeedButton read FButton;
    property Glyph: TBitmap read GetGlyph write SetGlyph;
    property NumGlyphs: integer read GetNumGlyphs write SetNumGlyphs;
    property KeepAspect: boolean read FKeepAspect write FKeepAspect default false;
    property OnBtnClick: TNotifyEvent read FOnBtnClick write SetOnBtnClick;
  published
    { published declarations }
  end;

  TCustomDBPopupEdit = class(TCustomPopupEdit)
  private
    { private declarations }
    FDataLink: TFieldDataLink;
    function GetDataField: string;
    function GetDataSource: TDataSource;
    function GetReadOnly: Boolean;
    procedure SetDataField(const Value: string);
    procedure SetDataSource(Value: TDataSource);
    procedure SetReadOnly(Value: Boolean);
    procedure SetFocused(Value: Boolean);
    procedure DataChange(Sender: TObject);
    procedure EditingChange(Sender: TObject);
    procedure UpdateData(Sender: TObject); { called when control changes }
    procedure WMCut(var Message: TMessage); message WM_CUT;
    procedure WMPaste(var Message: TMessage); message WM_PASTE;
    procedure CMEnter(var Message: TCMEnter); message CM_ENTER;
    procedure CMExit(var Message: TCMExit); message CM_EXIT;
    procedure CMFontChanged(var Message: TMessage); message CM_FONTCHANGED;
  protected
    { protected declarations }
    procedure Change; override;
    function EditCanModify: Boolean; override;
    procedure KeyDown(var Key: Word; Shift: TShiftState); override;
    procedure KeyPress(var Key: Char); override;
    procedure Notification(AComponent: TComponent;
      Operation: TOperation); override;
    procedure Reset; override;
  public
    { public declarations }
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    property DataField: string read GetDataField write SetDataField;
    property DataSource: TDataSource read GetDataSource write SetDataSource;
    property ReadOnly: Boolean read GetReadOnly write SetReadOnly default False;
  published
    { published declarations }
  end;

  TPopupEdit = class(TCustomPopupEdit)
  private
    { private declarations }
  protected
    { protected declarations }
  public
    { public declarations }
  published
    { published declarations }
    property AutoSelect;
    property AutoSize;
    property BorderStyle;
    property Color;
    property Ctl3D;
    property EditMask;
    property Font;
    property Glyph;
    property NumGlyphs;
    property KeepAspect;
    property ParentColor;
    property ParentCtl3D;
    property ParentFont;
    property ParentShowHint;
    property ReadOnly;
    property ShowHint;
    property Text;
    property TabOrder;
    property OnBtnClick;
    property OnChange;
    property OnClick;
    property OnDblClick;
    property OnEnter;
    property OnExit;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property DragCursor;
    property DragMode;
  end;

  TDBPopupEdit = class(TCustomDBPopupEdit)
  private
    { private declarations }
  protected
    { protected declarations }
  public
    { public declarations }
  published
    { public declarations }
    property AutoSelect;
    property AutoSize;
    property BorderStyle;
    property Color;
    property Ctl3D;
    property DataField;
    property DataSource;
    property Font;
    property Glyph;
    property KeepAspect;
    property NumGlyphs;
    property ParentColor;
    property ParentCtl3D;
    property ParentFont;
    property ParentShowHint;
    property ReadOnly;
    property TabOrder;
    property ShowHint;
    property OnBtnClick;
    property OnChange;
    property OnClick;
    property OnDblClick;
    property OnEnter;
    property OnExit;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property DragCursor;
    property DragMode;
  end;

  TDatePopupEdit = class(TCustomPopupEdit)
  private
    { private declarations }
    FPopup: TCustomPopup;
    procedure SetPopup(APopup: TCustomPopup);
  protected
    { protected declarations }
    procedure BtnClick(Sender: TObject); override;
  public
    { public declarations }
    constructor Create(AOwner: TComponent); override;
  published
    { published declarations }
    property Popup: TCustomPopup read FPopup write SetPopup;
    property AutoSelect;
    property AutoSize;
    property BorderStyle;
    property Color;
    property Ctl3D;
    property EditMask;
    property Font;
    property Glyph;
    property NumGlyphs;
    property KeepAspect;
    property ParentColor;
    property ParentCtl3D;
    property ParentFont;
    property ParentShowHint;
    property ReadOnly;
    property ShowHint;
    property Text;
    property TabOrder;
    property OnChange;
    property OnClick;
    property OnDblClick;
    property OnEnter;
    property OnExit;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property DragCursor;
    property DragMode;
  end;

TDBDatePopupEdit = class(TCustomDBPopupEdit)
  private
    { private declarations }
    FPopup: TCustomPopup;
    procedure SetPopup(APopup: TCustomPopup);
  protected
    { protected declarations }
    procedure BtnClick(Sender: TObject); override;
  public
    { public declarations }
    constructor Create(AOwner: TComponent); override;
  published
    { public declarations }
    property Popup: TCustomPopup read FPopup write SetPopup;
    property AutoSelect;
    property AutoSize;
    property BorderStyle;
    property Color;
    property Ctl3D;
    property DataField;
    property DataSource;
    property Font;
    property Glyph;
    property NumGlyphs;
    property KeepAspect;
    property ParentColor;
    property ParentCtl3D;
    property ParentFont;
    property ParentShowHint;
    property ReadOnly;
    property ShowHint;
    property TabOrder;
    property OnChange;
    property OnClick;
    property OnDblClick;
    property OnEnter;
    property OnExit;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property DragCursor;
    property DragMode;
  end;

procedure Register;

implementation

{$IFDEF WIN32}
  {$R POPEDITS.D32}
  {$R PEDIT32.RES}
{$ELSE}
  {$R POPEDITS.D16}
  {$R PEDIT16.RES}
{$ENDIF}

{ create component }
constructor TCustomPopupEdit.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);

  FButton := TSpeedButton.Create(Self);
  FButton.Width := 21;
  FButton.Height := 17;
  FButton.Visible := True;
  FButton.Glyph.Handle := LoadBitmap(HInstance, 'DARROW');
  FButton.NumGlyphs := 1;
  FButton.Parent := Self;
  FButton.Cursor := crArrow;
  FButton.OnClick := BtnClick;

  FKeepAspect := false;
  Width := 113;
  CalcTextMargin;
end;

{ destroy component }
destructor TCustomPopupEdit.Destroy;
begin
  { free button instance }
  FButton.Free;
  inherited Destroy;
end;

{ set component Params }
procedure TCustomPopupEdit.CreateParams(var Params: TCreateParams);
begin
  inherited CreateParams(Params);
  {  Params.Style := Params.Style and not WS_BORDER;  }
  Params.Style := Params.Style or ES_MULTILINE or WS_CLIPCHILDREN;
end;

{ CreateWnd procedure }
procedure TCustomPopupEdit.CreateWnd;
begin
  inherited CreateWnd;
  SetEditRect;
end;

function TCustomPopupEdit.GetGlyph: TBitMap;
begin
  Result := FButton.Glyph;
end;

procedure TCustomPopupEdit.SetGlyph(AGlyph: TBitmap);
begin
  FButton.Glyph.Assign(AGlyph);
end;

function TCustomPopupEdit.GetNumGlyphs: integer;
begin
  Result := FButton.NumGlyphs;
end;

procedure TCustomPopupEdit.SetNumGlyphs(Value: integer);
begin
  FButton.NumGlyphs := Value;
end;

procedure TCustomPopupEdit.SetOnBtnClick(Value: TNotifyEvent);
begin
  FOnBtnClick := Value;
end;

procedure TCustomPopupEdit.BtnClick(Sender: TObject);
begin
  if Assigned(FOnBtnClick) then FOnBtnClick(Self);
end;

{ CalcTextMargin method }
procedure TCustomPopupEdit.CalcTextMargin;
var
  DC: HDC;
  SaveFont: HFont;
  I: Integer;
  SysMetrics, Metrics: TTextMetric;
begin
  DC := GetDC(0);
  GetTextMetrics(DC, SysMetrics);
  SaveFont := SelectObject(DC, Font.Handle);
  GetTextMetrics(DC, Metrics);
  SelectObject(DC, SaveFont);
  ReleaseDC(0, DC);
  I := SysMetrics.tmHeight;
  if I > Metrics.tmHeight then I := Metrics.tmHeight;
  FTextMargin := I div 4;
end;

{ SetEditRect procedure }
procedure TCustomPopupEdit.SetEditRect;
var
  Loc: TRect;
begin
  SendMessage(Handle, EM_GETRECT, 0, LongInt(@Loc));
  Loc.Bottom := ClientHeight;
  Loc.Right := ClientWidth - FButton.Width - 2;
  Loc.Top := 0;
  Loc.Left := 0;
  SendMessage(Handle, EM_SETRECTNP, 0, LongInt(@Loc));
end;

{ windows WMSize method }
procedure TCustomPopupEdit.WMSize(var Message: TWMSize);
var
  MinHeight: Integer;
begin
  inherited;
  MinHeight := GetMinHeight;
  if Height < MinHeight then
    Height := MinHeight
  else if FButton <> nil then
  begin
    if FKeepAspect then FButton.Width := Height-(Height div 5);
    {$IFDEF WIN32}
    FButton.SetBounds(Width - FButton.Width-4, 0, FButton.Width, Height-4);
    {$ELSE}
    FButton.SetBounds(Width - FButton.Width, 0, FButton.Width, Height);
    {$ENDIF}
    SetEditRect;
  end;
end;

{ calculate minimum height }
function TCustomPopupEdit.GetMinHeight: Integer;
var
  DC: HDC;
  SaveFont: HFont;
  I: Integer;
  SysMetrics, Metrics: TTextMetric;
begin
  DC := GetDC(0);
  GetTextMetrics(DC, SysMetrics);
  SaveFont := SelectObject(DC, Font.Handle);
  GetTextMetrics(DC, Metrics);
  SelectObject(DC, SaveFont);
  ReleaseDC(0, DC);
  I := SysMetrics.tmHeight;
  if I > Metrics.tmHeight then I := Metrics.tmHeight;
  Result := Metrics.tmHeight + I div 4 + GetSystemMetrics(SM_CYBORDER) * 4 + 2;
end;

{ TCustomDBPopupEdit }
constructor TCustomDBPopupEdit.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  inherited ReadOnly := True;

  FDataLink := TFieldDataLink.Create;
  FDataLink.Control := Self;
  FDataLink.OnDataChange := DataChange;
  FDataLink.OnEditingChange := EditingChange;
  FDataLink.OnUpdateData := UpdateData;
end;


{ destroy component }
destructor TCustomDBPopupEdit.Destroy;
begin
  FDataLink.Free;
  FDataLink := nil;
  inherited Destroy;
end;

procedure TCustomDBPopupEdit.Change;
begin
  FDataLink.Modified;
  inherited Change;
end;

procedure TCustomDBPopupEdit.Notification(AComponent: TComponent;
  Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if (Operation = opRemove) and (FDataLink <> nil) and
    (AComponent = DataSource) then DataSource := nil;
end;

procedure TCustomDBPopupEdit.Reset;
begin
  FDataLink.Reset;
  SelectAll;
end;

procedure TCustomDBPopupEdit.KeyDown(var Key: Word; Shift: TShiftState);
begin
  inherited KeyDown(Key, Shift);
  if (Key = VK_DELETE) or ((Key = VK_INSERT) and (ssShift in Shift)) then
    FDataLink.Edit;
end;

procedure TCustomDBPopupEdit.KeyPress(var Key: Char);
begin
  inherited KeyPress(Key);
  if (Key in [#32..#255]) and (FDataLink.Field <> nil) and
    not FDataLink.Field.IsValidChar(Key) then
  begin
    MessageBeep(0);
    Key := #0;
  end;
  case Key of
    ^H, ^V, ^X, #32..#255:
      FDataLink.Edit;
    #27:
      begin
        FDataLink.Reset;
        SelectAll;
        Key := #0;
      end;
  end;
end;

procedure TCustomDBPopupEdit.WMPaste(var Message: TMessage);
begin
  FDataLink.Edit;
  inherited;
end;

procedure TCustomDBPopupEdit.WMCut(var Message: TMessage);
begin
  FDataLink.Edit;
  inherited;
end;

procedure TCustomDBPopupEdit.CMEnter(var Message: TCMEnter);
begin
  SetFocused(True);
  inherited;
end;

procedure TCustomDBPopupEdit.CMExit(var Message: TCMExit);
begin
  try
    FDataLink.UpdateRecord;
  except
    SelectAll;
    SetFocus;
    raise;
  end;
  SetFocused(False);
  SetCursor(0);
  DoExit;
end;

procedure TCustomDBPopupEdit.CMFontChanged(var Message: TMessage);
begin
  inherited;
  CalcTextMargin;
end;

function TCustomDBPopupEdit.EditCanModify: Boolean;
begin
  Result := FDataLink.Edit;
end;

procedure TCustomDBPopupEdit.SetFocused(Value: Boolean);
begin
  if FFocused <> Value then
  begin
    FFocused := Value;
    if (FAlignment <> taLeftJustify) and not IsMasked then Invalidate;
    FDataLink.Reset;
  end;
end;

{ GetDataSource method }
function TCustomDBPopupEdit.GetDataSource: TDataSource;
begin
  Result := FDataLink.DataSource;
end;

{ SetDataSource method }
procedure TCustomDBPopupEdit.SetDataSource(Value: TDataSource);
begin
  FDataLink.DataSource := Value;
end;

{ GetDataField method }
function TCustomDBPopupEdit.GetDataField: string;
begin
  Result := FDataLink.FieldName;
end;

{ SetDataField method }
procedure TCustomDBPopupEdit.SetDataField(const Value: string);
begin
  FDataLink.FieldName := Value;
end;

{ GetReadOnly method }
function TCustomDBPopupEdit.GetReadOnly: Boolean;
begin
  Result := FDataLink.ReadOnly;
end;

{ SetReadOnly method }
procedure TCustomDBPopupEdit.SetReadOnly(Value: Boolean);
begin
  FDataLink.ReadOnly := Value;
end;

procedure TCustomDBPopupEdit.DataChange(Sender: TObject);
begin
  if FDataLink.Field <> nil then
  begin
    if FAlignment <> FDataLink.Field.Alignment then
    begin
      EditText := '';  {forces update}
      FAlignment := FDataLink.Field.Alignment;
    end;
    EditMask := FDataLink.Field.EditMask;
    if FDataLink.Field.DataType = ftString then
      MaxLength := FDataLink.Field.Size else
      MaxLength := 0;
    if FFocused and FDataLink.CanModify then
      Text := FDataLink.Field.Text
    else
      EditText := FDataLink.Field.DisplayText;
  end else
  begin
    FAlignment := taLeftJustify;
    EditMask := '';
    MaxLength := 0;
    if csDesigning in ComponentState then
      EditText := Name else
      EditText := '';
  end;
end;

procedure TCustomDBPopupEdit.EditingChange(Sender: TObject);
begin
  inherited ReadOnly := not FDataLink.Editing;
end;

procedure TCustomDBPopupEdit.UpdateData(Sender: TObject);
begin
  ValidateEdit;
  FDataLink.Field.Text := Text;
end;

{ TDatePopup }
constructor TDatePopupEdit.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FButton.Glyph.Handle := LoadBitmap(HInstance, 'CALEDIT');
  Text := DateToStr(Date);
end;

procedure TDatePopupEdit.BtnClick(Sender: TObject);
var
  ADate: TDateTime;
begin
  if StrToDate(Text) <> 0 then
    ADate := StrToDate(Text)
  else
    ADate := Date;
  if FPopup.Execute(ADate) then
    Text := DateToStr(ADate);
end;

procedure TDatePopupEdit.SetPopup(APopup: TCustomPopup);
begin
  FPopup := APopup;
  try
    FPopup.AlignSource := Self;
  except
    FPopup.AlignSource := nil;
  end;
end;

{ TDBDatePopup }
constructor TDBDatePopupEdit.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FButton.Glyph.Handle := LoadBitmap(HInstance, 'CALEDIT');
end;

procedure TDBDatePopupEdit.BtnClick(Sender: TObject);
var
  ADate: TDateTime;
begin
  if FDataLink.Field <> nil then
  begin
    if FDataLink.Field.AsDateTime <> 0 then
      ADate := FDataLink.Field.AsDateTime
    else
      ADate := Date;
    if FPopup.Execute(ADate) then
    begin
      FDataLink.Edit;
      FDataLink.Field.AsDateTime := ADate;
    end;
  end;
end;

procedure TDBDatePopupEdit.SetPopup(APopup: TCustomPopup);
begin
  FPopup := APopup;
  try
    FPopup.AlignSource := Self;
  except
    FPopup.AlignSource := nil;
  end;
end;

{ register component on Misc page }
procedure Register;
begin
  RegisterComponents('HomeGrown', [TPopupEdit]);
  RegisterComponents('Data Controls', [TDBPopupEdit]);
  RegisterComponents('HomeGrown', [TDatePopupEdit]);
  RegisterComponents('Data Controls', [TDBDatePopupEdit]);
end;

end.